/* weedlog.c
 * Contains general stuff needed to run the program.
 */

#include "weedlog.h"

int socktypes = 0, bequiet = 0;

/* go_background();
 * Sends the program into the background (ie daemon mode).
 * Returns the new process id.
 */
int go_background() {
  int /*fd = -1,*/ fs = -1;

  switch((fs = fork())) {
    case -1:
      return (-1);
    case 0:
      break;
    default:
      exit(0);
  }

  if(setsid() == -1)
    return(-1);

  (void)chdir("/");
/*
  if((fd = open("/dev/null", O_RDWR, 0)) != -1) {
fprintf(stderr, "opened /dev/null withno errors\n");
    (void)dup2(fd, STDIN_FILENO);
fprintf(stderr, "dup2 STDIN_FILENO\n");
    (void)dup2(fd, STDOUT_FILENO);
fprintf(stderr, "dup2 STDOUT_FILENO\n");
    (void)dup2(fd, STDERR_FILENO);
fprintf(stderr, "du2 STDERR_FILENO\n");
    if(fd > 2)
      (void)close(fd);
  }*/
  return fs;

}

int showhelp(char *name) {
  fprintf(stdout, "weedlogd version %s\n", VERSION);
  fprintf(stdout, "Synopsis:\n");
#if !defined(__GLIBC__) || (__GLIBC__ < 2)
  fprintf(stdout, "  %s [-F <file>] [-I] [-T] [-U] [-b] [-d <#>] [-f] [-h] [-v[v...]] [-q]\n", name);
  fprintf(stdout, "    OR\n");
  fprintf(stdout, "  %s -[F][I][T][U][b][d][f][h][v[v...]][q] <file> <debug #>\n", name);
#else
  fprintf(stdout, "  %s [-I] [-T] [-U] [-d <#>] [-h] [-v[v...]] [-q]\n", name);
  fprintf(stdout, "    OR\n");
  fprintf(stdout, "  %s -[I][T][U][d][h][v[v...]][q] <debug #>\n", name);
#endif
  fprintf(stdout, "Command Line Options:\n");
#if !defined(__GLIBC__) || (__GLIBC__ < 2)
  fprintf(stdout, "  -F <file>\t: Use <file> instead of syslog when in Background/Daemon mode.\n");
#endif
  fprintf(stdout, "  -I\t\t: Watch for ICMP packets.\n");
  fprintf(stdout, "  -T\t\t: Watch for TCP packets.\n");
  fprintf(stdout, "  -U\t\t: Watch for UDP packets.\n");
#if !defined(__GLIBC__) || (__GLIBC__ < 2)
  fprintf(stdout, "  -b\t\t: Use Background/Daemon mode.\n");
#endif
  fprintf(stdout, "  -d #\t\t: Use debug mode level #.\n");
  fprintf(stdout, "\t\t- available debug modelevels: 5\n");
#if !defined(__GLIBC__) || (__GLIBC__ < 2)
  fprintf(stdout, "  -f\t\t: Use Foreground/Console mode.\n");
#endif
  fprintf(stdout, "  -h\t\t: This help message.\n");
  fprintf(stdout, "  -v\t\t: Set verbose level.\n");
  fprintf(stdout, "\t\t- Use once for each level, effective 3 times.\n");
  fprintf(stdout, "  -q\t\t: Be quiet when starting up (good for starting at boot).\n");
#if !defined(__GLIBC__) || (__GLIBC__ < 2)
  fprintf(stdout, "\nExamples:\n");
  fprintf(stdout, "  To verbosely watch TCP in the foreground (console):\n");
  fprintf(stdout, "    \'%s -vfvTv\' OR \'%s -f -T -v -v -v\'\n", name, name);
  fprintf(stdout, "  To verbosely watch ICMP, TCP and UDP with debuging in the"
      "background and send the output to a file instead of syslog:\n");
  fprintf(stdout, "    \'%s -vFvdIUTv /some/file 5\' OR \'%s -F /some/file -I "
      "-U -T -d 5 -v -v -v\'\n", name, name);
  fprintf(stdout, "\nNotes:\n");
  fprintf(stdout, "  - The options -f and -F may be used together to print to "
      "both stdout and the given filename at the same time. You cannot send to"
      " a file and to syslog at the same time (if running in the background).\n");
#else
  fprintf(stdout, "\nExamples:\n");
  fprintf(stdout, "  To verbosely watch TCP in the foreground (console):\n");
  fprintf(stdout, "    \'%s -vvTv\' OR \'%s -T -v -v -v\'\n", name, name);
  fprintf(stdout, "  To verbosely watch ICMP, TCP and UDP with debuging:\n");
  fprintf(stdout, "    \'%s -vvdIUTv 5\' OR \'%s -I -U -T -d 5 -v -v -v\'\n",
      name, name);
  fprintf(stdout, "\nNote:\n");
  fprintf(stdout, "    This appears to have been compiled under glibc version 2"
      " or newer. All output modes, aside from directly to console, have been "
      "disabled due to major inconcistencies I have not yet been able to work "
      "out. Feel free to try and fix them if you desire. Or simply recompile "
      "this application under libc5 for all features.\n");
#endif
  exit(0);
  return 0;
}

/* parse_args(argc, argv);
 * Parses the command line arguments.
 */
int parse_args(int argc, char *argv[]) {
  register const char *s;
  int n = 1;

  while(argv[n]) {
    s = argv[n];
    if(*s == '-') {
      s++;
      while(*s) {
#if !defined(__GLIBC__) || (__GLIBC__ < 2)
	if(*s == 'F') {
	  register const char *filepath;
	  char *fname;
	  n++;
	  filepath = argv[n];
	  if(!filepath) goto NEXT;
	  if((fname = (char*)malloc(PATH_MAX)) < 0) {
	    print_log("malloc: %s\n", strerror(errno));
	    goto NEXT;
	  }
	  if(*filepath == '/') {
	    snprintf(fname, PATH_MAX, "%s", filepath);
	  } else {
	    getcwd(fname, PATH_MAX);
	    snprintf(fname, PATH_MAX, "%s/%s", fname, filepath);
	  }
	  if((outfile = open(fname, O_WRONLY|O_CREAT|O_APPEND, S_IRUSR|S_IWUSR|S_IRGRP)) < 0) {
	    print_log("error: could not open %s (%s)\n", fname, strerror(errno));
	  }
	  if(fname)
	    free(fname);
	  goto NEXT;
	}
#endif
	if(*s == 'I') {
	  SetICMP(socktypes);
	  goto NEXT;
	}
	if(*s == 'T') {
	  SetTCP(socktypes);
	  goto NEXT;
	}
	if(*s == 'U') {
	  SetUDP(socktypes);
	  goto NEXT;
	}
	if(*s == 'b') {
	  daemon_mode = 1;
	  goto NEXT;
	}
	if(*s == 'd') {
	  n++;
	  debug = atoi(argv[n]);
	  goto NEXT;
	}
	if(*s == 'f') {
	  console_mode = 1;
	  goto NEXT;
	}
	if(*s == 'h') {
	  showhelp(argv[0]);
	  goto NEXT;
	}
	if(*s == 'v') {
	  verbose++;
	  goto NEXT;
	}
	if(*s == 'q') {
	  bequiet++;
	  goto NEXT;
	}
/*
	if(*s == '') {
	  goto NEXT;
	}
*/
NEXT:
	s++;
      }
    }
    n++;
  }

#if !defined(__GLIBC__) || (__GLIBC__ < 2)
  if(daemon_mode == 1 && console_mode == 1)
    console_mode = 0;
  if(daemon_mode == 0 && console_mode == 0)
    daemon_mode = 1;
#else
  if(!bequiet)
    printb_log(FALSE, "WARNING: You appear to be running this under glibc "
	"version 2 or newer. All output modes, aside from directly to console,"
	" have been disabled due to major inconcistencies I have not yet been "
	"able to work out. Feel free to try and fix them if you desire. Or "
	"simply recompile this application under libc5 for all features.\n");
  daemon_mode = 0;
  console_mode = 1;
#endif
  return 0;
}

int main(int argc, char *argv[]) {
  int bg;
  time_t t;

  progname = argv[0];
  wlpid = getpid();
  logmode = 0;
  verbose = 0;

  parse_args(argc, argv);

  setuid(0);
  if(geteuid() != 0)
    print_log("init: This program requires root privledges. (%s)\n",
	strerror(errno));

  if(!bequiet) {
    t = time(NULL);
    print_log("init: %s started at %s", progname, ctime(&t));
  }

  if(daemon_mode) {
    if(!bequiet)
      printb_log(FALSE, "init: Entering background mode.\n");
    if(!outfile) {
      char *syslogname;
      if((syslogname = (char*)malloc(PATH_MAX)) < 0) {
	print_log("malloc: %s", strerror(errno));
      }
      sprintf(syslogname, "%s[%i]", progname, wlpid);
      openlog(syslogname, 0, LOG_DAEMON);
      logmode = LOGMODE_SYSL;
      if(debug)
	print_log("Testing syslog status under name %s\n", syslogname);
//      if(syslogname) free(syslogname);
    } else {
      logmode = LOGMODE_FILE;
    }
    if((bg = go_background()) < 0) {
      print_log_err("error: fork() failed (%s), exiting.\n", strerror(errno));
      return 1;
    } else {
      wlpid = getpid();
      if(!bequiet)
	print_log("init: Background mode entered. PID changed to %i.\n", wlpid);
      if(logmode == LOGMODE_SYSL) {
	char *syslogname;
	if((syslogname = (char*)malloc(PATH_MAX)) < 0) {
	  print_log("malloc: %s", strerror(errno));
	}
	sprintf(syslogname, "%s[%i]", progname, wlpid);
	closelog();
	openlog(syslogname, 0, LOG_DAEMON);
	logmode = LOGMODE_SYSL;
	if(debug)
	  print_log("Testing syslog status under name %s\n", syslogname);
//	if(syslogname) free(syslogname);
      }
    }
  } else wlpid = getpid();

  if(outfile && !daemon_mode)
    logmode = LOGMODE_FFIL;

  setup_sockets(socktypes);
  if(!socketlist) {
    printb_log(FALSE, "error: No sockets to watch, exiting.\n");
    return 1;
  }

  for(;;) {
    if(debug >= 5)
      printb_log(FALSE, "debug: Checking for packets\n");
    check_for_packet();
    if(debug >= 5)
      printb_log(FALSE, "debug: Sleeping.\n");
//    usleep(1000000);
    usleep(100);
  }

  return 0;
}

